/*
  Copyright (C) 2010 DJ Delorie <dj@redhat.com>

  This file is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation; either version 2, or (at your option) any later
  version.
  
  This file is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with this file; see the file COPYING3.  If not see
  <http://www.gnu.org/licenses/>.
*/

#include <string.h>
#include "sfr_r8c33c.h"
#include "tty.h"

/*****************************************************************************/

typedef struct {
  char const * const name;
  void (*func)(void);
  char const *data;
} FileSystem;

static FileSystem commands[];

/*****************************************************************************/

extern unsigned int rand;


static char const hexc[] = "0123456789abcdef";

tty_hex4 (int h)
{
  tty_putc (hexc[(h>>12) & 15]);
  tty_putc (hexc[(h>>8) & 15]);
  tty_putc (hexc[(h>>4) & 15]);
  tty_putc (hexc[(h>>0) & 15]);
}

static void
hex(int v)
{
  tty_putc(hexc[v>>4]);
  tty_putc(hexc[v&15]);
}

static void
tty_int (unsigned int v, int w)
{
  static const unsigned int factors[] = { 10000, 1000, 100, 10, 1 };
  char buf[6];
  int i;
  char z = ' ';

  buf[5] = 0;

  for (i=0; i<5; i++)
    {
      unsigned int d = v / factors[i];
      v %= factors[i];

      if (d || i == 4)
	{
	  buf[i] = '0' + d;
	  if (w < 5-i)
	    w = 5-i;
	  z = '0';
	}
      else
	buf[i] = z;
    }
  tty_puts (buf+5-w);
}

/*****************************************************************************/

volatile int tics = 0;
volatile int up_ms = 0;
volatile int up_sec = 0;
volatile int up_min = 0;
volatile int up_hour = 0;

/* Timer RA handler */
void __attribute__((interrupt))
timer_ra_handler (void)
{
  tics ++;
  if (++up_ms == 1000)
    {
      up_ms = 0;
      if (++up_sec == 60)
	{
	  up_sec = 0;
	  if (++up_min == 60)
	    {
	      up_min = 0;
	      up_hour ++;
	    }
	}
    }
}

timer_init ()
{
  tracr = 0;
  tramr = 0;
  trapre = 127;
  tra = 143;
  traioc = 0;

  traic = 5;
  tstart_tracr = 1;
}

void
wait_ms (int ms)
{
  int now = tics;
  while ((unsigned int)(tics - now) < ms)
    ;
}

/*****************************************************************************/

char argc;
char *argv[10];
char buf[80];

/* Show is:
   0 - print asterisks
   1 - print actual text
   2 - print actual text, tab completion on command[]
*/

char *
tty_gets (int show)
{
  int i = 0;
  while (1)
    {
      char c = tty_getc ();

      if (c == 8 || c == 127)
	{
	  if (i > 0)
	    {
	      tty_puts("\b \b");
	      i--;
	    }
	  else
	    tty_putc('\a');
	}
      else if (c == 10 || c == 13)
	{
	  tty_putc('\n');
	  buf[i] = 0;
	  return buf;
	}
      else if (c == 9)
	{
	  int count = 0, ci;
	  int next_c;
	  for (ci=0; commands[ci].name; ci++)
	    if (commands[ci].func
		&& memcmp (commands[ci].name, buf, i) == 0)
	      {
		count ++;
		next_c = ci;
	      }
	  if (count == 1)
	    {
	      while (commands[next_c].name[i])
		{
		  tty_putc (commands[next_c].name[i]);
		  buf[i] = commands[next_c].name[i];
		  i ++;
		}
	      tty_putc (' ');
	      buf[i++] = ' ';
	    }
	  else
	    tty_putc('\a');
	}
      else if (c == 4 && i == 0)
	{
	  argc = -1;
	  return 0;
	}
      else if (isascii (c) && i < 79)
	{
	  buf[i++] = c;
	  if (show)
	    tty_putc (c);
	  else
	    tty_putc('*');
	}
    }
}

int
login ()
{
  int valid_account;

  tty_puts ("\n\033[32musername: \033[0m");

  if (tty_gets (1) == 0)
    return;
  valid_account = (strcmp (buf, "gr8c") == 0);

  tty_puts ("\033[32mpassword: \033[0m");

  if (tty_gets (0) == 0)
    return;
  if (strcmp (buf, "devcon")
      || !valid_account)
    {
      tty_puts("Invalid username or password\n");
      wait_ms (2000);
      return 0;
    }
  tty_puts("\nWelcome!\n");
  return 1;
}

/*****************************************************************************/

static void
mail()
{
  tty_puts("No mail.\n");
}

static void
twodig (int i)
{
  if (i > 99)
    {
      tty_putc ('0' + i/100);
      i %= 100;
      
    }
  tty_putc ('0' + i/10);
  tty_putc ('0' + i%10);
}

static void
uptime ()
{
  tty_puts ("up ");
  if (up_hour)
    {
      twodig (up_hour);
      tty_putc(':');
    }
  twodig (up_min);
  tty_putc(':');
  twodig (up_sec);
  tty_puts (", 1 user, load average: 0.01, 0.00, 0.00\n");
}

static void
leds ()
{
  int i;

  if (argc == 1)
    {
      tty_puts("leds set to");
      for (i=0; i<4; i++)
	printf(" %d", get_led (i));
      tty_putc('\n');
    }
  else
    {
      for (i=1; i<=4 && argv[i]; i++)
	set_led (i-1, atoi (argv[i]));
    }
}

static void
rgb ()
{
  int l;
  int i;
  int r, g, b;

  if (argc == 1)
    {
      get_rgb (&r, &g, &b);
      printf ("rgb set to %d %d %d\n", r, g, b);
    }
  else
    {
      set_rgb (atoi (argv[1]), atoi (argv[2]), atoi (argv[3]));
    }
}

static void
emacs ()
{
  tty_puts("emacs: not enough memory\n");
}

static void
uname ()
{
  tty_puts("FakeOS 1.0 R8C/33c\n");
}

static void
ls (void)
{
  int i;
  if (argc > 1
      && argv[1][0] == '-'
      && strchr (argv[1], 'l'))
    {
      for (i=0; commands[i].name; i++)
	{
	  if (commands[i].func)
	    tty_puts("---x     0 ");
	  else
	    {
	      tty_puts("-r-- ");
	      tty_int (strlen (commands[i].data), 5);
	      tty_puts(" ");
	    }
	  
	  tty_puts(commands[i].name);
	  if (commands[i].func && commands[i].data)
	    {
	      int j;
	      for (j=strlen(commands[i].name); j<14; j++)
		tty_putc(' ');
	      tty_putc('(');
	      tty_puts(commands[i].data);
	      tty_putc(')');
	    }
	  tty_putc('\n');
	}
      return;
    }
  for (i=0; commands[i].name; i++)
    {
      tty_puts(commands[i].name);
      tty_putc('\n');
    }
}

static void
shutdown ()
{
  tty_puts("Broadcase message from gr8c...\n");
  tty_puts("\n");
  tty_puts("This gR8C board is shutting down NOW!!!\n");
  wait_ms (2000);
  tty_puts("\nThis gR8C board is halted.\nPress any key to reboot...\n");
  while (!tty_ready ())
    {
      if (up_ms > 500)
	{
	  set_leds(0x0f);
	  set_rgb (255, 0, 0);
	}
      else
	{
	  set_leds(0x00);
	  set_rgb (0, 0, 0);
	}
    }
  tty_getc ();
  prcr = ~0;
  pm0 = 0xff; /* software reset */
}

static void
do_exit ()
{
}

static void
cat ()
{
  if (argc > 1)
    {
      int i;
      for (i=0; commands[i].name; i++)
	if (strcmp (commands[i].name, argv[1]) == 0)
	  {
	    if (! commands[i].data)
	      {
		tty_puts(argv[1]);
		tty_puts(": permission denied.\n");
		return;
	      }
	    tty_puts(commands[i].data);
	    return;
	  }
      tty_puts(argv[1]);
      tty_puts(": no such file.\n");
      return;
    }
  while (1)
    {
      argc == 0;
      if (tty_gets (1) == 0)
	return;
      tty_puts (buf);
      tty_putc('\n');
    }
}

/* The "world" is a 5x5 torus of rooms.  We start at 00.  There's a
   pit somewhere, and maybe the corpse of the last visitor.  In 20
   turns, your flashlight dies, and three turns later, you get eaten
   by a Grue.  Not much of a game :-)

   20 21 22 23 24
   15 16 17 18 19
   10 11 12 13 14
   05 06 07 08 09
   00 01 02 03 04

*/

static char const * const words[] = {
  /* a */ "You are in a",
  /* b */ " dark",
  /* c */ " twisty",
  /* d */ " twisting",
  /* e */ " twisty",
  /* f */ " twisted",
  /* g */ " maze",
  /* h */ " labyrinth",
  /* i */ " of",
  /* j */ " little",
  /* k */ " passages, all different.\n"
};
static char const * const phrases[] = {
  "abgik",      /* You are in a dark maze of passages, all different. */
  "abgick",     /* You are in a dark maze of twisty passages, all different. */
  "abgijk",     /* You are in a dark maze of little passages, all different. */
  "abgicjk",    /* You are in a dark maze of twisty little passages, all different. */
  "abgijck",    /* You are in a dark maze of little twisty passages, all different. */
  "abcgik",     /* You are in a dark twisty maze of passages, all different. */
  "abcgijk",    /* You are in a dark twisty maze of little passages, all different. */
  "abjgik",     /* You are in a dark little maze of passages, all different. */
  "abjgick",    /* You are in a dark little maze of twisty passages, all different. */
  "abcjgik",    /* You are in a dark twisty little maze of passages, all different. */
  "abjcgik"     /* You are in a dark little twisty maze of passages, all different. */
};

static char corpse = -1;
static char trap;
static char rooms[25];

static void
adventure ()
{
  int i, t, m;
  int turn = 0;
  int me = 0;
  int lastme = -1;
  const char *msg;
  int nsew = 0;

  /* Three bits for word variants, then mod 11. */
  for (i=0; i<25; i++)
    rooms[i] = -1;

  trap = rand % 25 + 1;
  tty_puts("\nPicoAdventure 1.0\n\n");
  while (1)
    {
      if (lastme != me)
	{
	  if (me == corpse)
	    tty_puts("There is a corpse here.\n");
	  if (rooms[me] == -1)
	    rooms[me] = rand % (11 * 8);

	  t = (rooms[me] >> 1) & 3;
	  m = rooms[me] & 1;
	  msg = phrases[rooms[me] >> 3];
	  for (i=0; msg[i]; i++)
	    {
	      int w = msg[i];
	      if (msg[i] == 'c')
		w += t;
	      if (msg[i] == 'g')
		w += m;
	      if (msg[i] == 'b' && turn < 20)
		continue;
	      tty_puts(words[w - 'a']);
	    }

	  if (nsew < 3)
	    tty_puts("There are rooms to the n, s, e, and w.\n");
	  if (nsew == 3)
	    tty_puts("There are rooms... well, you get the idea.\n");
	  nsew ++;
	  lastme = me;
	}

      tty_puts("\033[34mYour command? \033[36m");
      msg = tty_gets(1);
      tty_puts("\033[0m");
      if (msg == 0)
	return;
      if (stricmp(buf, "n") == 0)
	me = (me + 5) % 25;
      else if (stricmp(buf, "s") == 0)
	me = (me + 20) % 25;
      else if (stricmp(buf, "e") == 0)
	me = (me + 1) % 25;
      else if (stricmp(buf, "w") == 0)
	me = (me + 24) % 25;
      else if (stricmp (buf, "quit") == 0
	       || stricmp (buf, "exit") == 0)
	{
	  tty_puts("Bye!\n");
	  return;
	}
      else if (strcmp (buf, "xyzzy") == 0)
	{
	  tty_puts("You are teleported to... what do you think I am, Linux?\n");
	}
      else if (stricmp (buf, "look") == 0)
	{
	  lastme = -1;
	}
      else if (stricmp (buf, "eat") == 0)
	{
	  switch (rand % 3)
	    {
	    case 0:
	      tty_puts ("Dust, the breakfast of champions.\n");
	      break;
	    case 1:
	      tty_puts ("You chew your fingernails.\n");
	      break;
	    case 2:
	      tty_puts ("You nibble on some pocket lint.\n");
	      break;
	    }
	}
      else
	{
	  tty_puts("Huh?\n");
	  continue;
	}
      turn ++;
      if (me == trap)
	{
	  tty_puts("It's a trap!  You die!\n");
	  corpse = me;
	  return;
	}
      if (turn == 20)
	tty_puts ("Your batteries run out.\n");
      if (turn == 23)
	{
	  tty_puts ("The Grue eats you.\n");
	  corpse = me;
	  return;
	}
    }
}

static unsigned char kmask[] = { 0, 1, 2, 3, 3, 2, 1, 0, 0};

static void
kitt ()
{
  unsigned int a, b, p;
  int i = -1;

  tty_puts("Vroom! Vroom!\n");
  tty_puts("press any key to continue...\n");

  while (!tty_ready())
    {
      i = (i+1) % 8;

      a = kmask[i];
      set_led (a, 255);

      for (p=0; p<130; p++)
	{
	  for (b=0; b<4; b++)
	    {
	      if (b != a)
		{
		  int l = get_led (b);
		  if (l > 192)
		    set_led (b, l-2);
		  else if (l > 0)
		    set_led (b, l-1);
		}
	    }
	  wait_ms (1);
	}
    }
  set_leds (0);
  set_rgb (0, 0, 0);
  tty_getc ();
}

static void
dance ()
{
  int i;
  tty_puts("Dancing...   ");
  while (!tty_ready ())
    {
      unsigned char c = (*(unsigned char *)(0x8000 + i)
	      ^ *(unsigned char *)(0x8500 + i)
	      ^ i);
      set_leds (c);
      set_rgb ((c & 0x10) ? 255 : 0,
	       (c & 0x20) ? 255 : 0,
	       (c & 0x40) ? 255 : 0);
      i = (i + 1) & 0xff;
      wait_ms (100);
      tty_putc (8);
      tty_putc (8);
      tty_putc ("\\|/-"[i&3]);
      tty_putc(' ');
    }
  tty_getc ();
  tty_putc('\n');
  set_leds (0);
  set_rgb (0, 0, 0);
}

static void
diag ()
{
  int i = 0;
  while (!tty_ready ())
    {
      set_leds (0);
      set_rgb (0, 0, 0);
      switch (i)
	{
	case 0:
	  set_led (0, 255);
	  break;
	case 1:
	  set_led (0, 128);
	  break;
	case 2:
	  set_led (1, 255);
	  break;
	case 3:
	  set_led (1, 128);
	  break;
	case 4:
	  set_led (2, 255);
	  break;
	case 5:
	  set_led (2, 128);
	  break;
	case 6:
	  set_led (3, 255);
	  break;
	case 7:
	  set_led (3, 128);
	  break;
	case 8:
	  set_rgb (255, 0, 0);
	  break;
	case 9:
	  set_rgb (128, 0, 0);
	  break;
	case 10:
	  set_rgb (0, 255, 0);
	  break;
	case 11:
	  set_rgb (0, 128, 0);
	  break;
	case 12:
	  set_rgb (0, 0, 255);
	  break;
	case 13:
	  set_rgb (0, 0, 128);
	  break;
	case 14:
	  set_led (0, 255);
	  set_led (1, 255);
	  set_led (2, 255);
	  set_led (3, 255);
	  set_rgb (255, 255, 255);
	  break;
	case 15:
	  set_led (0, 128);
	  set_led (1, 128);
	  set_led (2, 128);
	  set_led (3, 128);
	  set_rgb (128, 128, 128);
	  break;
	}
      wait_ms (700);
      i = (i + 1) % 16;
    }
  tty_getc ();
  set_leds (0);
  set_rgb (0, 0, 0);
}

extern char _bssend[];
extern char _stack[];
extern char _datastart[];
extern char _romdatacopysize[];
extern char _romdatastart[];
extern char start[];

static void
mem ()
{
  char tos[1];
  unsigned memfree;
  unsigned memtotal;
  unsigned romtotal;
  unsigned romfree;
  char *rp;

  rp = _bssend;

  while (rp < tos && *rp == 0x11)
    rp ++;

  memfree = rp - _bssend;
  memtotal = _stack - _datastart;
  romtotal = (char *)(0xffffU) - start + 1;
  romfree = (char *)(0xffffU) - _romdatastart - (int)&_romdatacopysize;

  memtotal = (memtotal + 7) & ~7;

  tty_int (memtotal, 5); tty_puts(" bytes total RAM\n");
  tty_int (memfree, 5); tty_puts(" bytes free RAM\n");
  tty_int (memtotal - memfree, 5); tty_puts(" bytes used RAM\n");
  tty_int (_bssend - _datastart, 5); tty_puts(" bytes variables in RAM\n");
  tty_int (_stack - tos, 5); tty_puts(" bytes cur stack in RAM\n");
  tty_int (_stack - rp, 5); tty_puts(" bytes max stack in RAM\n");


  tty_int (romtotal, 5); tty_puts(" bytes total ROM\n");
  tty_int (romfree, 5); tty_puts(" bytes free ROM\n");
  tty_int (romtotal - romfree, 5); tty_puts(" bytes used ROM\n");


}

static void
dump ()
{
  int i, j;
  unsigned st, en;
  st = (unsigned)_datastart & ~0xff;
  en = ((unsigned)&i + 255) & ~0xff;

  for (i=st; i < en; i+= 32)
    {
      tty_hex4 (i);
      tty_puts(" : ");
      for (j=0; j<32; j++)
	{
	  tty_putc(' ');
	  switch (j) {
	  case 8:
	  case 24: tty_puts("\033[36m"); break;
	  case 16: tty_puts("\033[0m"); break;
	  }
	  hex(*(unsigned char *)(i+j));
	}
      tty_puts("\033[0m\n");
    }
}

static void
hello ()
{
  tty_puts("Hello, world\n");
}

static void
echo ()
{
  int i;
  for (i=1; i<argc; i++)
    {
      if (i > 1)
	tty_putc(' ');
      tty_puts(argv[i]);
    }
  tty_putc('\n');
}

static void
fill_ram ()
{
  char i;
  char *rp = _bssend;
  while (rp < &i)
    *rp++ = 0x11;
}

static char const * const gcc_version[] = {
#include "gcc-version.h"
  0
};

static void
gcc ()
{
  if (argc == 1)
    {
      tty_puts("gcc: no input files\n");
    }
  else if (strcmp (argv[1], "-v") == 0)
    {
      int i;
      for (i=0; gcc_version[i]; i++)
	{
	  tty_puts(gcc_version[i]);
	  tty_putc('\n');
	}
    }
  else
    {
      tty_puts("gcc: ");
      tty_puts(argv[1]);
      tty_puts(": No such file or directory\n");
    }
}

extern unsigned char sourceball[];
extern unsigned char sourceball_end[];

char const readme_text[] = "Welcome to FakeOS 1.0!!!\n"
  "This is a tiny fake os-like app just to keep you entertained.\n"
  "Is it working?\n";

static void
uuencode (unsigned char *data, int length)
{
  int i;
  tty_putc (' ' + length);
  while (length > 0)
    {
      unsigned long v = ((unsigned long)data[0] << 16) | ((unsigned long)data[1] << 8) | (unsigned long)data[2];
      for (i=18; i>=0; i-=6)
	{
	  char c = v >> i;
	  c &= 0x3f;
	  c += 32;
	  tty_putc (c == 32 ? '`' : c);
	}
      length -= 3;
      data += 3;
    }
  tty_putc('\n');
}

static void
sources ()
{
  int key;
  unsigned char *data, *dend;
  int length;

  printf("This will print a UUENCODED copy of fakeos.tar.gz.\n");
  printf("Press Y if this is OK, any other key to skip.  Press 'q' for a quick\n");
  printf("test file.  Use your terminal program's capture feature to save the text\n");
  printf("to a file, then uudecode it.\n");

  key = tty_getc ();
  if (key != 'y' && key != 'Y' && key != 'q' && key != 'Q')
    return;


  if (key == 'q' || key == 'Q')
    {
      tty_puts ("begin 664 README.txt\n");
      data = (unsigned char *)readme_text;
      dend = data + strlen (data);
    }
  else
    {
      tty_puts ("begin 664 fakeos.tar.gz\n");
      data = sourceball;
      dend = sourceball_end;
    }

  while (data < dend)
    {
      length = dend - data;
      if (length > 45)
	length = 45;
      uuencode (data, length);
      data += length;
    }
  tty_puts("`\nend\n\n");
}

extern void micromind();

static FileSystem commands[] = {
  { "COPYRIGHT", 0,
    "Copyright (C) 2010 DJ Delorie <dj@delorie.com>\n"
    "Distributed under the terms of the GNU General Public License v2\n" },
  { "README", 0, readme_text },
  { "adventure", adventure, "tiny text adventure game" },
  { "cat", cat, "print the contents of a file, or whatever you type" },
  { "core", 0, "" },
  { "dance", dance, "make the LEDs dance" },
  { "diag", diag, "LED diagnostics" },
  { "dump",  dump, "dump RAM contents" },
  { "echo", echo, "prints whatever you type after it" },
  { "emacs", emacs, "full featured text editor" },
  { "exit", do_exit, "log back out" },
  { "gcc", gcc, "C compiler" },
  { "hello", hello, "typical first application" },
  { "kitt", kitt, "red panning eyeball" },
  { "leds", leds, "read or write the LEDs manually" },
  { "ls", ls, "list all files" },
  { "mail", mail, "check your mail" },
  { "mem", mem, "see how much memory is left" },
  { "micromind", micromind, "guess the four colors" },
  { "rgb", rgb, "read or write the RGB LED manually" },
  { "shutdown", shutdown, "shut down the gR8C board" },
  { "sources", sources, "transfer the sources for this app to your PC" },
  { "uname", uname, "check the running OS name" },
  { "uptime", uptime, "see how long the gR8C board has been running" },
  { 0, 0 }
};


int
command ()
{
  int i;

  tty_puts("\n\033[34m$\033[36m ");
  if (tty_gets (1) == 0)
    {
      tty_puts("\033[0mexit\n");
      return 0;
    }
  tty_puts("\033[0m");

  argc = 0;
  i = 0;
  while (1)
    {
      while (buf[i] && isspace (buf[i]))
	buf[i++] = 0;
      if (buf[i] == 0)
	break;
      argv[argc] = buf + i;
      argc ++;
      while (buf[i] && ! isspace (buf[i]))
	i++;
    }
  if (argc == 0)
    return 1;

  if (strcmp (argv[0], "exit") == 0)
    return 0;

  for (i=0; commands[i].name; i++)
    if (strcmp (commands[i].name, argv[0]) == 0)
      {
	commands[i].func();
	return 1;
      }

  tty_puts(buf);
  tty_puts(": command not found\n");
  return 1;
}


main()
{
  p1 = 0;
  p2 = 1;
  pd1 = 0x0f;
  pd2 = 0x07;

  fill_ram ();

  timer_init ();

  tty_init ();

  tty_puts("\ngR8C FakeOS\n");

  while (1)
    {
      int tries = 0;
      while (!login())
	if (++tries == 5)
	  printf("here's a hint: username is gr8c, password is devcon.\n");

      while (command())
	;
    }
}
